%{

/**
 * parser.y -- SPL parser specification
 *
 * All modifications demanded by task 4 (building up an abstract syntax tree) are now included here.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
 
/* please note that the following includes must be in this order, as sadly the .h files do not 
 * include themselves what they need */
#include "common.h"
#include "utils.h"
#include "symbol.h"
#include "absyn.h"
#include "scanner.h"
#include "parser.h"

/**
 * global variable where our parser will output the program tree
 */
Absyn *progTree;

%}

%union {
  NoVal noVal;
  IntVal intVal;
  StringVal stringVal;
  Absyn *node;
}

%token	<noVal>		ARRAY ELSE IF OF PROC REF TYPE VAR WHILE
%token	<noVal>		LPAREN RPAREN LBRACK RBRACK LCURL RCURL
%token	<noVal>		EQ NE LT LE GT GE ASGN COLON COMMA SEMIC
%token	<noVal>		PLUS MINUS STAR SLASH
%token	<stringVal>	IDENT
%token	<intVal>	INTLIT

%type	<node>		type
%type	<node>		global_declarations global_declaration
%type	<node>		type_declaration
%type	<node>		procedure_declaration
%type	<node>		params non_empty_params parameter
%type	<node>		variable_declarations variable_declaration
%type	<node>		statements statement
%type	<node>		compound_statement
%type	<node>		assign_statement
%type	<node>		if_statement
%type	<node>		while_statement
%type	<node>		call_statement
%type	<node>		args non_empty_args
%type	<node>		expr add_expr mul_expr
%type	<node>		unary_expr primary_expr
%type	<node>		variable

%start			program


%%


program			: global_declarations
			  {
			    progTree = $1;
			  }
			;

type			: IDENT
			  {
			    $$ = newNametype($1.line, symbol($1.val));
			  }
			| ARRAY LBRACK INTLIT RBRACK OF type
			  {
			    $$ = newArraytype($1.line, $3.val, $6);
			  }
			;

global_declarations	: /* empty */
			  {
			    $$ = NULL;
			  }
			| global_declaration global_declarations
			  {
			    $$ = newList($1->line, $1, $2);
			  }
			;

global_declaration	: type_declaration
			  {
			    $$ = $1;
			  }
			| procedure_declaration
			  {
			    $$ = $1;
			  }
			;

type_declaration	: TYPE IDENT EQ type SEMIC
			  {
			    $$ = newTypedecl($1.line, symbol($2.val), $4);
			  }
			;

procedure_declaration	: PROC IDENT LPAREN params RPAREN
			  LCURL variable_declarations statements RCURL
			  {
			    $$ = newProcdecl($1.line, symbol($2.val),
			                     $4, $7, $8);
			  }
			;

params			: /* empty */
			  {
			    $$ = NULL;
			  }
			| non_empty_params
			  {
			    $$ = $1;
			  }
			;

non_empty_params	: parameter
			  {
			    $$ = newList($1->line, $1, NULL);
			  }
			| parameter COMMA non_empty_params
			  {
			    $$ = newList($1->line, $1, $3);
			  }
			;

parameter		: IDENT COLON type
			  {
			    $$ = newParam($1.line, symbol($1.val), $3, FALSE);
			  }
			| REF IDENT COLON type
			  {
			    $$ = newParam($2.line, symbol($2.val), $4, TRUE);
			  }
			;

variable_declarations	: /* empty */
			  {
			    $$ = NULL;
			  }
			| variable_declaration variable_declarations
			  {
			    $$ = newList($1->line, $1, $2);
			  }
			;

variable_declaration	: VAR IDENT COLON type SEMIC
			  {
			    $$ = newVardecl($1.line, symbol($2.val), $4);
			  }
			;

statements		: /* empty */
			  {
			    $$ = NULL;
			  }
			| statement statements
			  {
			    /* $1 may be an empty compound statement */
			    if ($1 != NULL) {
			      $$ = newList($1->line, $1, $2);
			    } else {
			      $$ = $2;
			    }
			  }
			;

statement		: compound_statement
			  {
			    $$ = $1;
			  }
			| assign_statement
			  {
			    $$ = $1;
			  }
			| if_statement
			  {
			    $$ = $1;
			  }
			| while_statement
			  {
			    $$ = $1;
			  }
			| call_statement
			  {
			    $$ = $1;
			  }
			;

compound_statement	: LCURL statements RCURL
			  {
			    $$ = $2;
			  }
			;

assign_statement	: variable ASGN expr SEMIC
			  {
			    $$ = newAssign($2.line, $1, $3);
			  }
			;

if_statement		: IF LPAREN expr RPAREN statement
			  {
			    $$ = newIf($1.line, $3, $5, NULL);
			  }
			| IF LPAREN expr RPAREN statement ELSE statement
			  {
			    $$ = newIf($1.line, $3, $5, $7);
			  }
			;

while_statement		: WHILE LPAREN expr RPAREN statement
			  {
			    $$ = newWhile($1.line, $3, $5);
			  }
			;

call_statement		: IDENT LPAREN args RPAREN SEMIC
			  {
			    $$ = newCall($1.line, symbol($1.val), $3);
			  }
			;

args			: /* empty */
			  {
			    $$ = NULL;
			  }
			| non_empty_args
			  {
			    $$ = $1;
			  }
			;

non_empty_args		: expr
			  {
			    $$ = newList($1->line, $1, NULL);
			  }
			| expr COMMA non_empty_args
			  {
			    $$ = newList($1->line, $1, $3);
			  }
			;

expr			: add_expr
			  {
			    $$ = $1;
			  }
			| add_expr EQ add_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_EQ, $1, $3);
			  }
			| add_expr NE add_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_NE, $1, $3);
			  }
			| add_expr LT add_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_LT, $1, $3);
			  }
			| add_expr LE add_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_LE, $1, $3);
			  }
			| add_expr GT add_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_GT, $1, $3);
			  }
			| add_expr GE add_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_GE, $1, $3);
			  }
			;

add_expr		: mul_expr
			  {
			    $$ = $1;
			  }
			| add_expr PLUS mul_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_ADD, $1, $3);
			  }
			| add_expr MINUS mul_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_SUB, $1, $3);
			  }
			;

mul_expr		: unary_expr
			  {
			    $$ = $1;
			  }
			| mul_expr STAR unary_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_MUL, $1, $3);
			  }
			| mul_expr SLASH unary_expr
			  {
			    $$ = newOp($2.line, ABSYN_OP_DIV, $1, $3);
			  }
			;

unary_expr		: primary_expr
			  {
			    $$ = $1;
			  }
			| MINUS primary_expr
			  {
			    $$ = newOp($1.line, ABSYN_OP_SUB,
			               newInt($1.line, 0), $2);
			  }
			;

primary_expr		: variable
			  {
			    $$ = newVar($1->line, $1);
			  }
			| INTLIT
			  {
			    $$ = newInt($1.line, $1.val);
			  }
			| LPAREN expr RPAREN
			  {
			    $$ = $2;
			  }
			;

variable		: IDENT
			  {
			    $$ = newSimplevar($1.line, symbol($1.val));
			  }
			| variable LBRACK expr RBRACK
			  {
			    $$ = newArrayvar($1->line, $1, $3);
			  }
			;


%%


void yyerror(char *msg) {
  error("%s in line %d", msg, yylval.noVal.line);
}
