%{

/**
 * scanner.l -- SPL scanner specification
 * 
 * @author Matthias Ansorg
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

/* please note that the following includes must be in this order, as sadly the .h files do not 
 * include themselves what they need */
#include "common.h"
#include "utils.h"
#include "scanner.h"
#include "symbol.h"
#include "absyn.h"
#include "parser.tab.h"

static int lineNumber = 1;


/*
 * Return the value of given hex constant.
 * The constant may be of the form of either
 * 0x??? or ???H, where ??? is a sequence of
 * hex digits.
 *
 * I borrowed this from I http://cvs.sourceforge.net/viewcvs.py/asm48/asm48/lex.l?rev=1.6 .
 */
static int hex_const_value(const char *str)
{
  int value;

  if (str[0] == '0' && tolower(str[1]) == 'x') {
    sscanf(str + 2, "%x", &value);
  } else {
    value = 0;
    for (;;) {
      int ch = tolower(*str++);
      if (ch == 'h')
        break;
      value <<= 4;
      if (isdigit(ch))
        value += (ch - '0');
      else {
        assert(ch >= 'a' && ch <= 'f');
        value += (10 + (ch - 'a'));
      }
    }
  }
  return value;
}

%}

/* shortcuts for the regexps that match different kinds of int literals */
ASCIIINT  '.'
NEWLINEINT '\\n'
HEXINT    0x[A-Fa-f0-9]+
NORMALINT [0-9]+

%%

[\t ]+ { /* ignore withespace / nothing to do */
}

\/\/.* { /* ignore comments / nothing to do */
}

\n { /* @todo: Zeilennummer erhhen */
  ++lineNumber;
}

array {
  yylval.noVal.line = lineNumber;
  return ARRAY;
}

else {
  yylval.noVal.line = lineNumber;
  return ELSE;
}

if {
  yylval.noVal.line = lineNumber;
  return IF;
}

of { /* "array of ..." syntax */
  yylval.noVal.line = lineNumber;
  return OF;
}

proc {
  yylval.noVal.line = lineNumber;
  return PROC;
}

ref { /* marker for reference variables */
  yylval.noVal.line = lineNumber;
  return REF;
}

type {
  yylval.noVal.line = lineNumber;
  return TYPE;
}

var { /* marker for normal variables */
  yylval.noVal.line = lineNumber;
  return VAR;
}

while {
  yylval.noVal.line = lineNumber;
  return WHILE;
}

\( {
  yylval.noVal.line = lineNumber;
  return LPAREN;
}

\) {
  yylval.noVal.line = lineNumber;
  return RPAREN;
}

\[ {
  yylval.noVal.line = lineNumber;
  return LBRACK;
}

\] {
  yylval.noVal.line = lineNumber;
  return RBRACK;
}

\{ {
  yylval.noVal.line = lineNumber;
  return LCURL;
}

\} {
  yylval.noVal.line = lineNumber;
  return RCURL;
}

= {
  yylval.noVal.line = lineNumber;
  return EQ;
}

# {
  yylval.noVal.line = lineNumber;
  return NE;
}

\< {
  yylval.noVal.line = lineNumber;
  return LT;
}

\<= {
  yylval.noVal.line = lineNumber;
  return LE;
}

\> {
  yylval.noVal.line = lineNumber;
  return GT;
}

\>= {
  yylval.noVal.line = lineNumber;
  return GE;
}

:= {
  yylval.noVal.line = lineNumber;
  return ASGN;
}

: {
  yylval.noVal.line = lineNumber;
  return COLON;
}

, {
  yylval.noVal.line = lineNumber;
  return COMMA;
}

; {
  yylval.noVal.line = lineNumber;
  return SEMIC;
}

\+ {
  yylval.noVal.line = lineNumber;
  return PLUS;
}

- {
  yylval.noVal.line = lineNumber;
  return MINUS;
}

\* {
  yylval.noVal.line = lineNumber;
  return STAR;
}

\/ {
  yylval.noVal.line = lineNumber;
  return SLASH;
}

[a-zA-Z_][A-Za-z_0-9]* {
  yylval.stringVal.line = lineNumber;
  yylval.stringVal.line = lineNumber;
  yylval.stringVal.val = (char*)allocate(yyleng+1);
  strcpy(yylval.stringVal.val, yytext); // we must copy characters, not pointers
  return IDENT;
}

{NORMALINT} {
  yylval.intVal.line = lineNumber;
  yylval.intVal.val = atoi(yytext);
  return INTLIT;
}

{ASCIIINT} {
  yylval.intVal.line = lineNumber;
  yylval.intVal.val = yytext[0];
  return INTLIT;
}

{NEWLINEINT} {
  yylval.intVal.line = lineNumber;
  yylval.intVal.val = 0x0a;
  return INTLIT;
}

{HEXINT} {
  yylval.intVal.line = lineNumber;
  yylval.intVal.val = hex_const_value(yytext);;
  return INTLIT;
}

%%

int yywrap(void) {
  return 1;
}

/* functionality to print textual representation of tokens not needed for the parser */

void showToken(int token) {
  printf("zeile %i \t",lineNumber);
  switch (token) {
    case ARRAY:
      printf("array");
    break;
    case ELSE:
      printf("else");
    break;
    case IF:
      printf("if");
    break;
    case OF:
      printf("of");
    break;
    case PROC:
      printf("proc");
    break;
    case REF:
      printf("ref");
    break;
    case TYPE:
      printf("type");
    break;
    case VAR:
      printf("var");
    break;
    case WHILE:
      printf("while");
    break;
    case LPAREN:
      printf("lparen");
    break;
    case RPAREN:
      printf("rparen");
    break;
    case LBRACK:
      printf("lbrack");
    break;
    case RBRACK:
      printf("rbrack");
    break;
    case LCURL:
      printf("lcurl");
    break;
    case RCURL:
      printf("rcurl");
    break;
    case EQ:
      printf("eq");
    break;
    case NE:
      printf("ne");
    break;
    case LT:
      printf("lt");
    break;
    case LE:
      printf("le");
    break;
    case GT:
      printf("gt");
    break;
    case GE:
      printf("ge");
    break;
    case ASGN:
      printf("asgn");
    break;
    case COLON:
      printf("colon");
    break;
    case COMMA:
      printf("comma");
    break;
    case SEMIC:
      printf("semic");
    break;
    case PLUS:
      printf("plus");
    break;
    case MINUS:
      printf("minus");
    break;
    case STAR:
      printf("star");
    break;
    case SLASH:
      printf("slash");
    break;
    case IDENT:
      printf("ident '%s'",yylval.stringVal.val);
    break;
    case INTLIT:
      printf("intlit: %i", yylval.intVal.val);
    break;
    case 0:
      printf("EOF");
    break;
  }
  printf("\n");

}