/**
 * semant.c -- semantic analysis
 *
 * @todo There are approx. 7 test cases in negtests_sem/ where this semantic analysis does not 
 *   find the error. See messages of "make negtests_sem".
 * @todo Add further test cases with errors that must be found by this semantic analysis.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "utils.h"
#include "symbol.h"
#include "types.h"
#include "absyn.h"
#include "table.h"
#include "semant.h"


static boolean showLocalTables;

static Type *intType;
static Type *boolType;


/* forward declaration */
static Type *checkNode(Absyn *node, Table *symTable, int pass);


static Type *checkList(Absyn *node, Table *symTable, int pass) {
  while (node != NULL) {
    checkNode(node->u.listNode.head, symTable, pass);
    node = node->u.listNode.tail;
  }
  return NULL;
}


static Type *checkNametype(Absyn *node, Table *symTable, int pass) {
  Entry *entry;
  Type *type;

  // showTable(symTable);
  
  entry = lookup(symTable, node->u.nametypeNode.name);
  if (entry == NULL) {
    error("undefined type '%s' in line %d",
          symbolString(node->u.nametypeNode.name), node->line);
  }
  if (entry->kind != ENTRY_KIND_TYPE) {
    error("'%s' is not a type in line %d",
          symbolString(node->u.nametypeNode.name), node->line);
  }
  type = entry->u.typeEntry.type;
  return type;
}


static Type *checkArraytype(Absyn *node, Table *symTable, int pass) {
  Type *baseType;
  Type *type;

  /* the following checkNode() might branch to checkNametype() or checkArraytype() */
  baseType = checkNode(node->u.arraytypeNode.type, symTable, pass);
  
  type = newArrayType(node->u.arraytypeNode.size, baseType);
  return type;
}


static Type *checkTypedecl(Absyn *node, Table *symTable, int pass) {
  Type *type;
  Entry *entry;

  if (pass != 1) {
    return NULL; // do nothing
  }
  else {
    type = checkNode(node->u.typedeclNode.type, symTable, pass);
    entry = newTypeEntry(type);
    if (enter(symTable, node->u.typedeclNode.name, entry) == NULL) {
      error("redeclaration of '%s' in line %d",
            symbolString(node->u.typedeclNode.name), node->line);
    }
    return NULL;
  }
}


static Type *checkVardecl(Absyn *node, Table *symTable, int pass) {
  Entry *entry;
  Type *type;

  type = checkNode(node->u.vardeclNode.type, symTable, pass);
  entry = newVarEntry(type, FALSE);
  
  /* as enter() does not allow to enter the same _symbol_ twice in a symbol table, the following 
   * checks both if our variable collides with a previous variable declaration and with a 
   * parameter declaration. */
  if (enter(symTable, node->u.vardeclNode.name, entry) == NULL) {
    error("redeclaration of '%s' in line %d",
          symbolString(node->u.vardeclNode.name), node->line);
  }
  return NULL;
}

/** checks semantically one parameter of a procedure
 * 
 * While the parameters are what you write in your procedure's declaration (you see, a parameter 
 * is something that might assume different values), the arguments are what you write in 
 * your procedure's call.
 */
static Type *checkParam(Absyn *node, Table *symTable, int pass) {
  Type *type;
  boolean isRef;

  type = checkNode(node->u.paramNode.type, symTable, pass);
  isRef = node->u.paramNode.isRef;
  
  if (type->kind == TYPE_KIND_ARRAY && !isRef) {
    error("array parameter '%s' must be given as reference parameters in line %d",
          symbolString(node->u.paramNode.name), node->line);
  }
  return type;
}

/** checks semantically the parameters of a procedure
 * 
 * While the parameters are what you write in your procedure's declaration (you see, a parameter 
 * is something that might assume different values), the arguments are what you write in 
 * your procedure's call.
 */
static ParamTypes *checkParams(Absyn *node, Table *symTable, int pass) {
  Entry *entry;
  Type *type;
  
  /* this can only happen with an empty parameter list */
  if (node == NULL) {
    return NULL; // same as checkList(node, symTable, pass), but different pointer type
  }

  type = checkParam(node->u.listNode.head, symTable, pass);
  entry = newVarEntry(type, node->u.listNode.head->u.paramNode.isRef);

  if (enter(symTable, node->u.listNode.head->u.paramNode.name, entry) == NULL) {
    error("redeclaration of '%s' in line %d",
          symbolString(node->u.listNode.head->u.paramNode.name), node->line);
  }
  
  return newParamTypes(type, 
    node->u.listNode.head->u.paramNode.isRef, 
    checkParams(node->u.listNode.tail, symTable, pass)
  );
}


static Type *checkProcdecl(Absyn *node, Table *symTable, int pass) {
  Entry *procEntry;
  Table *localTable;
  Absyn *params;
  Absyn *param;
  ParamTypes *paramTypes;
  Type *paramType;
  boolean paramIsRef;
  Entry *paramEntry;

  if (pass == 1) {
    /* create local symbol table */
    localTable = newTable(symTable);
    
    /* enter parameters into local symbol table;
     * we know this is a list, but more specifically: a list of parameters; so we cannot simply 
     * call checkList() as this will neither do the spcific parameter checks nor can it branch to 
     * checkParams() as it does not know that the list is a parameters list */
    paramTypes = checkParams(node->u.procdeclNode.params, localTable, pass);
    
    /* build and enter proc declaration into symbol table */
    procEntry = newProcEntry(paramTypes, localTable); // paramTypes and localTable yet undetermined
    if (enter(symTable, node->u.procdeclNode.name, procEntry) == NULL) {
      error("redeclaration of '%s' in line %d",
            symbolString(node->u.procdeclNode.name), node->line);
    }
    
    /* enter local declarations into local symbol table */
    checkNode(node->u.procdeclNode.decls, localTable, pass);
    
    /* show symbol table if requested */
    if (showLocalTables) { // global variable
      showTable(localTable);
    }
  }
  else {
    /* get the proc declaration entered in pass 1 again */
    procEntry = lookup(symTable, node->u.procdeclNode.name);
    
    localTable = procEntry->u.procEntry.localTable;
    
    /* do semantic check on procedure body */
    checkNode(node->u.procdeclNode.body, localTable, pass);
  }
  return NULL;
}


static Type *checkAssign(Absyn *node, Table *symTable, int pass) {
  Type *varType;
  Type *exprType;

  varType = checkNode(node->u.assignNode.var, symTable, pass);
  exprType = checkNode(node->u.assignNode.expr, symTable, pass);
  
  /* the following compares pointers, equivalent to variable instances; this is correct as 
   * "same type" here means "same type entry in symbol table" */
  if (varType != exprType) {
    error("assignment combines different types in line %d", node->line);
  }
  
  /* The SPL language definition says: operations on an array as a whole are not allowed, except 
   * giving the array as argument to a reference parameter. Accordingly, the reference compiler 
   * requires int variables for assignments, so do we. */
  if (varType != intType) {
    error("assignment requires integer variable to the left in line %d", node->line);
  }
  /* no need here to check if node->u.assignNode.var is a variable. This was already done by 
   * checkNode() above, as the absyn tree guarantees that the left hand side element of an 
   * assignment is a simplevarNode or a arrayvarNode. */
  return NULL; // nobody needs the assignment type; there might even be no such thing in SPL
}


/**
 * @todo what is returned by checkNode() for boolean expression types? Correct usage here?
 */
static Type *checkIf(Absyn *node, Table *symTable, int pass) {
  Type *testType;

  testType = checkNode(node->u.ifNode.test, symTable, pass);
  if (testType != boolType) {
    error("if's test not of type boolean in line %d",
          node->line);
  }
  checkNode(node->u.ifNode.thenPart, symTable, pass);
  checkNode(node->u.ifNode.elsePart, symTable, pass); // might be NULL
  return NULL;
}


static Type *checkWhile(Absyn *node, Table *symTable, int pass) {
  Type *testType;

  testType = checkNode(node->u.whileNode.test, symTable, pass);
  if (testType != boolType) {
    error("while's test not of type boolean in line %d",
          node->line);
  }
  checkNode(node->u.whileNode.body, symTable, pass);
  return NULL;
}

/** checks semantically the arguments of a procedure call
 * 
 * While the parameters are what you write in your procedure's declaration (you see, a parameter 
 * is something that might assume different values), the arguments are what you write in 
 * your procedure's call.
 *
 * @todo line number information for the message "parameter and argument counts differ"
 */
static Type *checkArgs(Absyn *node, Table *symTable, int pass, ParamTypes *paramTypes) {
  Type *typeActual;
  Type *typeExpected;
  boolean isRef;

  if (node == NULL && paramTypes == NULL) {
    return NULL; // nothing left to do; same as checkList(node, symTable, pass)
  }
  else {
    if ((node == NULL && paramTypes != NULL) || (node != NULL && paramTypes == NULL)) {
      error("parameter and argument counts differ");
    }
  }
  
  // PRE: node != NULL && paramTypes != NULL
  typeActual = checkNode(node->u.listNode.head, symTable, pass);
  typeExpected = paramTypes->type;
  isRef = paramTypes->isRef;
  if (typeActual != typeExpected ) {
    error("parameter type mismatch in line %d", node->line);
  }
  if (isRef &&  node->u.listNode.head->type != ABSYN_VAR) {
    // printf("Debug :: checkArgs :: node->type is %d\n", node->type);
    error("for reference parameters, the argument must be a variable in line %d", node->line);
  }
  
  checkArgs(node->u.listNode.tail, symTable, pass, paramTypes->next);
  return NULL;
}
 
/**
 * @todo we do not know what's the node type of arguments / argument lists; correct 
 *   handling here anyway?
 */
static Type *checkCall(Absyn *node, Table *symTable, int pass) {
  ParamTypes *paramTypes;
  Entry* entry;
  
  /* check symbol table if symbol node->u.callNode.name is a defined procedure */
  entry = lookup(symTable, node->u.callNode.name);
  if (entry == NULL) {
    error("undefined procedure '%s' in line %d",
          symbolString(node->u.callNode.name), node->line);
  }
  if (entry->kind != ENTRY_KIND_PROC) {
    error("'%s' is not a procedure in line %d",
          symbolString(node->u.callNode.name), node->line);
  }

  /* check list of arguments */
  checkArgs(node->u.callNode.args, symTable, pass, entry->u.procEntry.paramTypes);

  return NULL;  
}


static Type *checkOp(Absyn *node, Table *symTable, int pass) {
  Type *leftType;
  Type *rightType;
  Type *type;

  leftType = checkNode(node->u.opNode.left, symTable, pass);
  rightType = checkNode(node->u.opNode.right, symTable, pass);
  if (leftType != rightType) {
    error("expression combines different types in line %d",
          node->line);
  }
  switch (node->u.opNode.op) {
    case ABSYN_OP_EQ:
    case ABSYN_OP_NE:
    case ABSYN_OP_LT:
    case ABSYN_OP_LE:
    case ABSYN_OP_GT:
    case ABSYN_OP_GE:
      if (leftType != intType) {
        error("comparison requires integer operands in line %d",
              node->line);
      }
      type = boolType;
      break;
    case ABSYN_OP_ADD:
    case ABSYN_OP_SUB:
    case ABSYN_OP_MUL:
    case ABSYN_OP_DIV:
      if (leftType != intType) {
        error("arithmetic operation requires integer operands in line %d",
              node->line);
      }
      type = intType;
      break;
    default:
      error("unknown op %d in checkOp", node->u.opNode.op);
  }
  
  return type;
}

/** checks a variable node semantically
 * 
 * Absyn's varNodes are just "container nodes" for Absyn's simplevarNodes and arrayvarNodes.
 * varNodes are used to provide a single interface for the different kind of variable nodes. So, 
 * one might on variable node type simply by "if (node->type == ABSYN_VAR)", instead of saying 
 * "if (node->type == ABSYN_SIMPLEVAR || node->type == ABSYN_ARRAYVAR)".
 * 
 * In situations where this common interface is not required, the varNode container is not used.
 */
static Type *checkVar(Absyn *node, Table *symTable, int pass) {
  Type *type;

  /* this might branch to checkSimplevar() or checkArrayvar() */
  type = checkNode(node->u.varNode.var, symTable, pass);
  return type;
}

/**
 * checks integer literals semantically, and returns their type "intType"
 */
static Type *checkInt(Absyn *node, Table *symTable, int pass) {
  Type *type;

  type = intType;
  return type;
}


static Type *checkSimplevar(Absyn *node, Table *symTable, int pass) {
  Entry *entry;
  Type *type;

  entry = lookup(symTable, node->u.simplevarNode.name);
  if (entry == NULL) {
    error("undefined variable '%s' in line %d",
          symbolString(node->u.simplevarNode.name), node->line);
  }
  if (entry->kind != ENTRY_KIND_VAR) {
    error("'%s' is not a variable in line %d",
          symbolString(node->u.simplevarNode.name), node->line);
  }
  type = entry->u.varEntry.type;
  return type;
}

/** sematic checking for array variables (e.g. foo[3])
 *
 * This does not include checking if the index is within the limits of the 
 * array type, as SPL language definition says: using an array index beyond 
 * these limits generates a runtime error (i.e., not a compile time error). 
 * Most times it will even be impossible to check the limits of an array 
 * index, as (dynamic) variable contents might be used to calculate it.
 *
 * You see, "arrayvar" is not the best name for what we check here. If you give a whole array 
 * as function argument, it appears as simplevar node in the abstract syntax. So we'd better 
 * say "indexedvar" instead of "arrayvar".
 */
static Type *checkArrayvar(Absyn *node, Table *symTable, int pass) {
  Type *varType;
  Type *indexType;
  
  /* check if variable is of array type */
  varType = checkNode(node->u.arrayvarNode.var, symTable, pass);
  if (varType->kind != TYPE_KIND_ARRAY) {
    error("indexed variable is not an array in line %d", node->line);
    /* we did not mention the name of node->u.arrayvarNode.var, as we do not know it: it might be 
     * another arrayvarNode */
  }
  
  /* check if node->u.arrayvarNode.index is of type intType */
  indexType = checkNode(node->u.arrayvarNode.index, symTable, pass);
  /* the following compares pointers, equivalent to variable instances; this is correct as 
   * "same type" here means "same type entry in symbol table" */
  if (indexType != intType) {
    error("array index not of int type in line %d", node->line);
  }
  
  // PRE: varType is of TYPE_KIND_ARRAY (tested above)
  return varType->u.arrayType.baseType; // type of the array's selected element, not the whole array
}

/** does the semantic check for a full abstract syntax tree
 *
 * This function is "hidden recursive": node type specific functions called herein will branch 
 * back to checkNode() if they encounter a yet unspecific sub-node.
 *  
 * @param node the abstract syntax tree to check
 * @param symTable the symbol table that belongs to this abstract syntax tree; it if filled 
 *   with the symbols while doing the semantic check.
 * @param pass the number of the checkNode() call for this abstract syntax tree; it may be 1 or 2; 
 *   the second pass was introduced to cope with forward references as appearing in recursive calls.
 * @return the type of the checked node, if applicable; else, NULL is returned, e.g. for variable 
 *   declarations and type declarations.
 */
static Type *checkNode(Absyn *node, Table *symTable, int pass) {
  if (node == NULL) {
    /* this can only happen with an empty list */
    return checkList(node, symTable, pass);
  }
  else {
    switch (node->type) {
      case ABSYN_LIST:
        return checkList(node, symTable, pass);
      case ABSYN_NAMETYPE:
        return checkNametype(node, symTable, pass);
      case ABSYN_ARRAYTYPE:
        return checkArraytype(node, symTable, pass);
      case ABSYN_TYPEDECL:
        return checkTypedecl(node, symTable, pass);
      case ABSYN_VARDECL:
        return checkVardecl(node, symTable, pass);
      case ABSYN_PARAM:
        return checkParam(node, symTable, pass);
      case ABSYN_PROCDECL:
        return checkProcdecl(node, symTable, pass);
      case ABSYN_ASSIGN:
        return checkAssign(node, symTable, pass);
      case ABSYN_IF:
        return checkIf(node, symTable, pass);
      case ABSYN_WHILE:
        return checkWhile(node, symTable, pass);
      case ABSYN_CALL:
        return checkCall(node, symTable, pass);
      case ABSYN_OP:
        return checkOp(node, symTable, pass);
      case ABSYN_VAR:
        return checkVar(node, symTable, pass);
      case ABSYN_INT:
        return checkInt(node, symTable, pass);
      case ABSYN_SIMPLEVAR:
        return checkSimplevar(node, symTable, pass);
      case ABSYN_ARRAYVAR:
        return checkArrayvar(node, symTable, pass);
      default:
        error("unknown node type %d in checkNode", node->type);
        /* not reached */
        return NULL;
    }
  }
}


static void enterPredefinedTypes(Table *table) {
  enter(table,
        symbol("int"),
        newTypeEntry(intType));

  /* SPL language definition says: SPL uses boolean values, but you can neither create boolean 
   * variables nor are there boolean literals. Therefore we do not enter the boolean type boolType 
   * into the global symbol table */
}


static void enterPredefinedProcs(Table *table) {
  Entry *procEntry;

  /* printi(i: int) */
  procEntry = newProcEntry(newParamTypes(intType, FALSE, NULL),
                           table);
  enter(table, symbol("printi"), procEntry);

  /* printc(i: int) */
  procEntry = newProcEntry(newParamTypes(intType, FALSE, NULL),
                           table);
  enter(table, symbol("printc"), procEntry);

  /* readi(ref i: int) */
  procEntry = newProcEntry(newParamTypes(intType, TRUE, NULL),
                           table);
  enter(table, symbol("readi"), procEntry);

  /* readc(ref i: int) */
  procEntry = newProcEntry(newParamTypes(intType, TRUE, NULL),
                           table);
  enter(table, symbol("readc"), procEntry);

  /* exit() */
  procEntry = newProcEntry(NULL,
                           table);
  enter(table, symbol("exit"), procEntry);

  /* time(ref i: int) */
  procEntry = newProcEntry(newParamTypes(intType, TRUE, NULL),
                           table);
  enter(table, symbol("time"), procEntry);

  /* clearAll(color: int) */
  procEntry = newProcEntry(newParamTypes(intType, FALSE, NULL),
                           table);
  enter(table, symbol("clearAll"), procEntry);

  /* setPixel(x: int, y: int, color: int) */
  procEntry = newProcEntry(newParamTypes(intType, FALSE,
                             newParamTypes(intType, FALSE,
                               newParamTypes(intType, FALSE, NULL))),
                           table);
  enter(table, symbol("setPixel"), procEntry);

  /* drawLine(x1: int, y1: int, x2: int, y2: int, color: int) */
  procEntry = newProcEntry(newParamTypes(intType, FALSE,
                             newParamTypes(intType, FALSE,
                               newParamTypes(intType, FALSE,
                                 newParamTypes(intType, FALSE,
                                   newParamTypes(intType, FALSE, NULL))))),
                           table);
  enter(table, symbol("drawLine"), procEntry);

  /* drawCircle(x0: int, y0: int, radius: int, color: int) */
  procEntry = newProcEntry(newParamTypes(intType, FALSE,
                             newParamTypes(intType, FALSE,
                               newParamTypes(intType, FALSE,
                                 newParamTypes(intType, FALSE, NULL)))),
                           table);
  enter(table, symbol("drawCircle"), procEntry);
}


void checkMain(Table *globalTable) {
  Entry *mainEntry;

  mainEntry = lookup(globalTable, symbol("main"));
  if (mainEntry == NULL) {
    error("procedure 'main' is missing");
  }
  if (mainEntry->kind != ENTRY_KIND_PROC) {
    error("'main' is not a procedure");
  }
  if (mainEntry->u.procEntry.paramTypes != NULL) {
    error("procedure 'main' must not have any parameters");
  }
}

/** 
 * one and only interface function to spl semantic analysis
 *
 * Just write it down and you're done with this stuff.
 *
 * @param program abstract syntax tree of the SPL program for which this 
 *   function will perform sematic checking.
 * @param showSymbolTables If set to TRUE, this function will 
 *   print the local symbol tables out. Otherwise, it will not do that. While a whole 
 *   SPL program has a global symbol table, each of its procedures has a local symbol 
 *   table additionally. For showing these symbol tables, the function "showTable()" is used.
 *   Showing the local symbol tables is no obligatory part of the compiler building homeworks, 
 *   though recommended for debugging purposes.
 * @return the global symbol table corresponding to parameter "program". It will contain all 
 *   symbols of this program and additionally all predefined types and procedures of SPL. It is 
 *   filled while doing the semantic checking within this function. You can reach the local symbol 
 *   tables via the procedure entries in the global symbol table.
 *
 */
Table *check(Absyn *program, boolean showSymbolTables) {
  Table *globalTable;

  /* store global flag to show local symbol tables; 
   * showLocalTables is a boolean variable that is global within this compilation unit */
  showLocalTables = showSymbolTables;
  
  /* generate built-in types into global variables;
   * these will be inserted into the global symbol table by enterPredefinedTypes() */
  intType = newPrimitiveType("int");
  boolType = newPrimitiveType("boolean");
  
  /* set up global symbol table */
  globalTable = newTable(NULL); /* NULL: global table has no upper level table */
  enterPredefinedTypes(globalTable);
  enterPredefinedProcs(globalTable);
  
  /* do semantic checks; to cope with forward references, we use two phases; 
   * forward references are introduced by recursion */
  checkNode(program, globalTable, 1);
  checkNode(program, globalTable, 2);
  
  /* check if "main" exists and is a procedure without parameters */
  checkMain(globalTable);
  
  /* return global symbol table */
  return globalTable;
}
