/*
 * table.c -- symbol table
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "utils.h"
#include "symbol.h"
#include "types.h"
#include "table.h"

/** creates and returns an symbol table entry representing a type entry
 * 
 * @param type struct reprensenting a type; create this with the 
 *   functions newPrimitiveType() or newArrayType().
 */
Entry *newTypeEntry(Type *type) {
  Entry *entry;

  entry = (Entry *) allocate(sizeof(Entry));
  entry->kind = ENTRY_KIND_TYPE;
  entry->u.typeEntry.type = type;
  return entry;
}

/** creates and returns an symbol table entry representing a variable entry
 * 
 * @param type struct reprensenting a type; create this with the 
 *   functions newPrimitiveType() or newArrayType().
 * @param idRef indicates if the variable is a reference
 */
Entry *newVarEntry(Type *type, boolean isRef) {
  Entry *entry;

  entry = (Entry *) allocate(sizeof(Entry));
  entry->kind = ENTRY_KIND_VAR;
  entry->u.varEntry.type = type;
  entry->u.varEntry.isRef = isRef;
  return entry;
}

/** creates and returns an symbol table entry representing a procedure entry
 * 
 * @param type struct reprensenting the procedure's formal parameters; create this with the 
 *   function newParamTypes().
 * @param localTable local symbol table of this procedure; the whole SPL program shares 
 *   a global symbol table, and each procedure has a local symbol table additionally; 
 *   you need to create this first /using newTable()) before you can call newProcEntry() 
 *   for a procedure.
 */
Entry *newProcEntry(ParamTypes *paramTypes, Table *localTable) {
  Entry *entry;

  entry = (Entry *) allocate(sizeof(Entry));
  entry->kind = ENTRY_KIND_PROC;
  entry->u.procEntry.paramTypes = paramTypes;
  entry->u.procEntry.localTable = localTable;
  return entry;
}

/** creates and returnes a new symbol table
 * 
 * You will need this to create the global symbol table of the whole SPL program and a local 
 * symbol table for each of its procedures.
 *
 * @param upperLevel points to the symbol table that is one level more global than that one you 
 *   want to create; for example, the global symbol table is in this relation to every symbol 
 *   tables of procedures.
 */
Table *newTable(Table *upperLevel) {
  Table *table;

  table = (Table *) allocate(sizeof(Table));
  table->bintree = NULL;
  table->upperLevel = upperLevel;
  return table;
}

/** enters one record ("line") into a symbol table
 * 
 * You will need this to put the entries you created with newVarEntry(), 
 * newTypeEntry(), newProcEntry() into the symbol table.
 *
 * @param table the symbol table you want to put your record into
 * @param symbol the symbol ("name", "identifier") representing an entry in your program
 * @param entry the entry represented in your SPL program by the parameter "symbol"
 * @return Entry Returns what was inserted into the symbol table: a Entry* variable if 
 *   your entry was successfully inserted; or NULL if your symbol was already found in the 
 *   symbol table.
 */
Entry *enter(Table *table, Symbol *symbol, Entry *entry) {
  unsigned key;
  Bintree *newtree;
  Bintree *current;
  Bintree *previous;

  key = symbolStamp(symbol);
  newtree = (Bintree *) allocate(sizeof(Bintree));
  newtree->symbol = symbol;
  newtree->key = key;
  newtree->entry = entry;
  newtree->left = NULL;
  newtree->right = NULL;
  if (table->bintree == NULL) {
    table->bintree = newtree;
  } else {
    current = table->bintree;
    while (1) {
      if (current->key == key) {
        /* symbol already in table */
        return NULL;
      }
      previous = current;
      if (current->key > key) {
        current = current->left;
      } else {
        current = current->right;
      }
      if (current == NULL) {
        if (previous->key > key) {
          previous->left = newtree;
        } else {
          previous->right = newtree;
        }
        break;
      }
    }
  }
  return entry;
}

/** looks up and returns an entry in a binary tree
 * 
 * @return The pointer to the entry if an entry with key "key" was found. Or 
 *   NULL, if such an entry was not found.
 */
static Entry *lookupBintree(Bintree *bintree, unsigned key) {
  while (bintree != NULL) {
    if (bintree->key == key) {
      return bintree->entry;
    }
    if (bintree->key > key) {
      bintree = bintree->left;
    } else {
      bintree = bintree->right;
    }
  }
  return NULL;
}

/** looks up an symbol in a symbol table and returns the corresponding entry
 *
 * @param table the table to look up the symbol in; this function will successively access 
 *   the more global symbol tables, if the entry was not found in this one.
 * @param symbol the symbol to look up in the symbol table
 * @return if the symbol "symbol" was found in the symbol table "table" or one of its parents, 
 *   returns the entry corresponding to this symbol; else, returns NULL.
 */
Entry *lookup(Table *table, Symbol *symbol) {
  unsigned key;
  Entry *entry;

  key = symbolStamp(symbol);
  while (table != NULL) {
    entry = lookupBintree(table->bintree, key);
    if (entry != NULL) {
      return entry;
    }
    table = table->upperLevel;
  }
  return NULL;
}


void showEntry(Entry *entry) {
  switch (entry->kind) {
    case ENTRY_KIND_TYPE:
      printf("type: ");
      showType(entry->u.typeEntry.type);
      break;
    case ENTRY_KIND_VAR:
      printf("var: ");
      if (entry->u.varEntry.isRef) {
        printf("ref ");
      }
      showType(entry->u.varEntry.type);
      break;
    case ENTRY_KIND_PROC:
      printf("proc: ");
      showParamTypes(entry->u.procEntry.paramTypes);
      break;
    default:
      error("unknown entry kind %d in showEntry", entry->kind);
  }
  printf("\n");
}


static void showBintree(Bintree *bintree) {
  if (bintree == NULL) {
    return;
  }
  showBintree(bintree->left);
  printf("  %-10s --> ", symbolString(bintree->symbol));
  showEntry(bintree->entry);
  showBintree(bintree->right);
}


void showTable(Table *table) {
  int level;

  level = 0;
  while (table != NULL) {
    printf("  level %d\n", level);
    showBintree(table->bintree);
    table = table->upperLevel;
    level++;
  }
}
